#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include "darray.h"

DWORD CDynamicArrayBase::dwPageSize;

CDynamicArrayBase::CDynamicArrayBase(DWORD dwS)
    : dwTopCommit(0), dwTopRequest(0), fCommit(TRUE)
{
    if ( dwPageSize == 0 ) {
        SYSTEM_INFO info;
        GetSystemInfo(&info);      // Inquire system page size.
        dwPageSize = info.dwPageSize;
    }
    dwSize = ToPages(dwS);
    dwBottomCommit = dwBottomRequest = ToBytes(dwSize);
    // Reserve the required pages:
    pBase = LPSTR(VirtualAlloc(NULL, dwBottomCommit,
                  MEM_RESERVE, PAGE_NOACCESS));
    fFail = pBase == NULL;
}

CDynamicArrayBase::~CDynamicArrayBase(void)
{
    // Release the pages:
    VirtualFree(LPVOID(pBase), 0, MEM_RELEASE);
}

INT CDynamicArrayBase::ExceptionHandler(DWORD dwCode,
        DWORD dwIndex)
{
    return ( fCommit && dwCode == EXCEPTION_ACCESS_VIOLATION
        && dwIndex < ToBytes(dwSize) &&
        Commit(dwIndex, dwIndex) )
        ? EXCEPTION_CONTINUE_EXECUTION
        : EXCEPTION_EXECUTE_HANDLER;
}

BOOL CDynamicArrayBase::Commit(DWORD dwFrom, DWORD dwTo)
{
    // Pointer to first byte to be committed.
    LPSTR pFrom = FromBase(dwFrom);
    // Pointer to last byte to be committed.
    LPSTR pTo = ToBase(dwTo)+dwPageSize;
    if ( pFrom < pTo && VirtualAlloc(pFrom, pTo-pFrom,
            MEM_COMMIT, PAGE_READWRITE) ) {
        if ( dwFrom < dwBottomRequest )
            dwBottomRequest = dwFrom;
        if ( dwTo > dwTopRequest )
            dwTopRequest = dwTo;
        dwFrom = ToElementsCorrected(pFrom-pBase);
        dwTo = ToElements(pTo-pBase-1);
        if ( dwFrom < dwBottomCommit )
            dwBottomCommit = dwFrom;
        if ( dwTo > dwTopCommit )
            dwTopCommit = dwTo;
        Init(dwFrom, dwTo);      // Initialize the range.
        return TRUE;
    }
    return FALSE;
}

VOID CDynamicArrayBase::Release(DWORD dwFrom, DWORD dwTo)
{
    LPSTR pFrom = FromBase(dwFrom)+((FromOffset(dwFrom) > 0)
        ? dwPageSize : 0);
    LPSTR pTo = ToBase(dwTo)+((ToOffset(dwTo) == PageMask())
        ? dwPageSize : 0);
    if ( pFrom < pTo ) {
        if ( dwTo >= dwBottomRequest
                && dwFrom <= dwBottomRequest )
            dwBottomRequest = dwTo+1;
        if ( dwFrom <= dwTopRequest && dwTo >= dwTopRequest )
            dwTopRequest = dwFrom-1;
        dwFrom = ToElements(pFrom-pBase);
        dwTo   = ToElementsCorrected(pTo-pBase-1);
        Cleanup(dwFrom, dwTo);			// Cleanup the range.
        if ( dwTo >= dwBottomCommit
                && dwFrom <= dwBottomCommit )
            dwBottomCommit = dwTo+1;
        if ( dwFrom <= dwTopCommit && dwTo >= dwTopCommit )
            dwTopCommit = dwFrom-1;
        // Decommit pages:
        VirtualFree(pFrom, pTo-pFrom, MEM_DECOMMIT);
    }
}
