/* mysort.c - a simple disk-based sort utility
 * Note: this is not a general sort utility, I only need it
 * to work for my particular data set!!
 */

#include <assert.h>
#include <stdio.h>
#include <windows.h>

int IsNT        = FALSE;
int RecSize;
int NRecs;

#define HEAP_SIZE   (200000)
//#define HEAP_SIZE   (2)

#define INPUT_CHUNK (500)

int WDJ_FormatErrorA(DWORD Error,
        char* Output, int Length)
    {
    return FormatMessageA(FORMAT_MESSAGE_IGNORE_INSERTS
        |FORMAT_MESSAGE_FROM_SYSTEM,
        NULL, Error,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
        Output, Length, NULL);
    }


class   TAsynchIO
    {
public:
    TAsynchIO(int BufferSize);
   ~TAsynchIO();
    int     StartRead(HANDLE File, DWORD Offset);
    char*   Buffer(int i);
    int     Wait();
private:
    DWORD       NBytesRead;
    int         NRecsRead;
    char*       _Buffer;
    int         BufferSize;
    HANDLE      FileHandle;
    HANDLE      EventHandle;
    OVERLAPPED  AsynchInfo;
    };
char*   TAsynchIO::Buffer(int i)
    {
    if(i < NRecsRead)
        return _Buffer + (i * RecSize);
    else
        return 0;
    }

TAsynchIO::TAsynchIO(int Size)
    {
    BufferSize  = Size;
    _Buffer     = (char*)malloc(INPUT_CHUNK*RecSize);
    assert(_Buffer != 0);
    EventHandle = CreateEvent(NULL, TRUE, FALSE, NULL);
    assert(EventHandle != 0);
    }
TAsynchIO::~TAsynchIO()
    {
    free(_Buffer);
    CloseHandle(EventHandle);
    }

int TAsynchIO::StartRead(HANDLE File, DWORD Offset)
    {
    FileHandle  = File;
    // initialize data structure
    memset(&AsynchInfo, 0, sizeof(AsynchInfo));
    AsynchInfo.Offset    = Offset;
    ResetEvent(EventHandle);
    AsynchInfo.hEvent   = EventHandle;
    NBytesRead          = 0;
    OVERLAPPED* Info    = IsNT ? &AsynchInfo : 0;
    BOOL Status         = ReadFile(File, _Buffer, BufferSize,
                            &NBytesRead,
                            Info
                            );
    if(Status == FALSE)
        {
        DWORD   Err = GetLastError();
        if(Err != ERROR_IO_PENDING)
            {
            char    Buffer[256];
            WDJ_FormatErrorA(Err, Buffer, sizeof(Buffer));
            fprintf(stderr, "Error %ld on ReadFile(%d): '%s'\n",
                Err, BufferSize, Buffer);
            return FALSE;
            }
        }
    return TRUE;
    }

// wait for the I/O to complete
int TAsynchIO::Wait()
    {
    if(IsNT)
        {
        GetOverlappedResult(FileHandle, &AsynchInfo, &NBytesRead, TRUE);
        }
    NRecsRead   = NBytesRead / RecSize;
    return NBytesRead;
    }


class   TRecordInput
    {
public:
    TRecordInput();
   ~TRecordInput();
    int Open(const char* Filename);
    int Close();
    int GetRecord(char* Record);
private:
    int         RecNum;
    char*       RecPtr;
    HANDLE      FileHandle;
    int         FileOfs;
    int         CurBufNum;
    int         CurBufOfs;
    int         CurBufCnt;
    TAsynchIO*  IO[2];
    };

TRecordInput::TRecordInput()
    : FileHandle(0), CurBufNum(0)
    {
    IO[0]   = new TAsynchIO(INPUT_CHUNK * RecSize);
    assert(IO[0] != NULL);
    IO[1]   = new TAsynchIO(INPUT_CHUNK * RecSize);
    assert(IO[1] != NULL);
    }
TRecordInput::~TRecordInput()
    {
    delete IO[0];
    delete IO[1];
    Close();
    }

int TRecordInput::Open(const char*Filename)
    {
    RecNum      = 0;
    RecPtr      = 0;
    CurBufNum   = 1;
    CurBufOfs   = (INPUT_CHUNK*RecSize);
    CurBufCnt   = INPUT_CHUNK;
SetLastError(0);
    FileHandle  = CreateFile(Filename, GENERIC_READ, 0,
            NULL, OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL
//            | FILE_FLAG_OVERLAPPED
//            | FILE_FLAG_SEQUENTIAL_SCAN
            , 0);
DWORD Err = GetLastError();
if(Err != 0)
    {
    char Buffer[256];
    WDJ_FormatErrorA(Err, Buffer, sizeof(Buffer));
    fprintf(stderr, "error on open is %ld '%s'\n", Err, Buffer);
    }

    if(FileHandle == (HANDLE)0)
        {
        assert(FALSE);
        return FALSE;
        }
            
    // spin off first asynchronous read
    if(IO[0]->StartRead(FileHandle, FileOfs) == FALSE)
        assert(FALSE);
    return TRUE;
    }

int TRecordInput::Close()
    {
    if(FileHandle != 0)
        {
        CloseHandle(FileHandle);
        FileHandle  = 0;
        }
    }

// Get next input record
int TRecordInput::GetRecord(char* Output)
    {
    // if we have exhausted this buffer
    if(CurBufCnt == INPUT_CHUNK)
        {
        // start a new read on this buffer
        IO[CurBufNum]->StartRead(FileHandle, FileOfs + (INPUT_CHUNK * RecSize));
        // switch to alternate buffer
        CurBufNum   = !CurBufNum;
        // wait for its I/O to complete
        IO[CurBufNum]->Wait();
        CurBufCnt   = 0;
        }
    // now we are set to perform the read
    char *Next = IO[CurBufNum]->Buffer(CurBufCnt);
    if(Next)
        {
        memcpy(Output, Next, RecSize);
        FileOfs += RecSize;
        ++CurBufCnt;
        }
    return Next != 0;
    }

void Usage(void)
    {
    fprintf(stderr,
        "Usage: mysort <filename> <RecordSize> <nrecords>\n");
    exit(EXIT_FAILURE);
    }


int Compare(const void*a, const void*b)
    {    return strncmp((const char*)a,(const char*)b,78);    }


void main(int argc, char** argv)
    {
    OSVERSIONINFO   Info;
    Info.dwOSVersionInfoSize    = sizeof(Info);
    GetVersionEx(&Info);
    IsNT    = (Info.dwPlatformId == VER_PLATFORM_WIN32_NT);

    if(argc != 4)
        Usage();

    RecSize = atoi(argv[2]);
    NRecs   = atoi(argv[3]);
    if(RecSize <= 0 || NRecs <= 0)
        Usage();
    fprintf(stderr, "RecSize=%d, NRecs=%d\n", RecSize, NRecs);

    TRecordInput    Input;
    if(!Input.Open(argv[1]))
        {
        fprintf(stderr, "Can't open '%s'\n", argv[1]);
        Usage();
        }
    char*   Heap    = (char*)malloc(HEAP_SIZE * RecSize);
    int Count = 0;
    int FileNum = 0;
    while(Count < NRecs)
        {
        // fill the heap
        for(int i=0; i < HEAP_SIZE; ++i)
            {
            if(!Input.GetRecord(Heap + (i*RecSize)))
                {
                NRecs   = Count;
                break;
                }
            else if(++Count >= NRecs)
                break;
            }
        if(i > 0)
            {
            qsort(Heap, i, RecSize, Compare);
            char FileName[256];
            sprintf(FileName, "E:T%04d.tmp", FileNum++);
            FILE *Out = fopen(FileName, "wb");
            fwrite(Heap, RecSize, i, Out);
            fclose(Out);
            }
        }
    // now do the merge
    TRecordInput** Inputs   = new TRecordInput*[FileNum];
    char *Buffer            = new char[RecSize * FileNum];
    char**Records           = new char*[FileNum];
    for(int i=0; i < FileNum; ++i)
        {
        Inputs[i]   = new TRecordInput;
        Records[i]  = Buffer + (i*RecSize);
        char FileName[256];
        sprintf(FileName, "E:T%04d.tmp", i);
        Inputs[i]->Open(FileName);
        if(!Inputs[i]->GetRecord(Records[i]))
            Records[i]  = 0;
        }
    FILE*Out    = fopen("F:sorted.dat", "wb");
    
    // while records left to process
    while(NRecs > 0)
        {
        int j;
        // while heap is not filled
        for(j = 0; j < HEAP_SIZE && NRecs > 0; ++j)
            {
            int iSmall;
            // find smallest available record
            char* Smallest = Records[iSmall=0];
            for(int k=1; k < FileNum; ++k)
                {
                if((Records[k] && !Smallest)
                || (Records[k] && strncmp(Records[k], Smallest, 78) < 0)
                  )
                    {
                    iSmall      = k;
                    Smallest    = Records[k];
                    }
                }
            if(Smallest == 0)
                NRecs   = 0;
            else
                {
                memcpy(Heap+(j*RecSize), Smallest, 80);
                --NRecs;
                if(!Inputs[iSmall]->GetRecord(Records[iSmall]))
                    Records[iSmall]  = 0;
                }
            }

        fwrite(Heap, RecSize, j, Out);
        }

    fclose(Out);

    exit(EXIT_SUCCESS);
    }

